# ----------------------------------------------------------------------------------------------------------------------
# - Package Imports -
# ----------------------------------------------------------------------------------------------------------------------
# General Packages
from __future__ import annotations
from dataclasses import dataclass
from typing import Literal, ClassVar

# Athena Packages

# Local Imports
from AthenaTwitchLib.irc.tags._tags import Conversion, Tags, TAG_TYPES

# ----------------------------------------------------------------------------------------------------------------------
# - Code -
# ----------------------------------------------------------------------------------------------------------------------
@dataclass(slots=True, frozen=True)
class TagsUSERNOTICE(Tags):
    """
    Class for Twitch IRC Tags, that are from the USERNOTICE message
    """
    badge_info:str=None
    badges:str=None
    color:str=None
    display_name:str=None
    emotes:str=None
    id:str=None
    login:str=None
    mod:bool=None
    msg_id:str=None
    room_id:str=None
    system_msg:str=None
    turbo:bool=None
    user_id:str=None
    user_type:Literal["", "admin", "global_mod", "staff"]=None

    msg_param_cumulative_months:str=None
    msg_param_displayName:str=None
    msg_param_login:str=None
    msg_param_months:str=None
    msg_param_promo_gift_total:str=None
    msg_param_promo_name:str=None
    msg_param_recipient_display_name:str=None
    msg_param_recipient_id:str=None
    msg_param_recipient_user_name:str=None
    msg_param_sender_login:str=None
    msg_param_sender_name:str=None
    msg_param_should_share_streak:str=None
    msg_param_streak_months:str=None
    msg_param_sub_plan:str=None
    msg_param_sub_plan_name:str=None
    msg_param_viewerCount:str=None
    msg_param_ritual_name:str=None
    msg_param_threshold:str=None
    msg_param_gift_months:str=None


    _tag_type:ClassVar[TAG_TYPES] = TAG_TYPES.USERNOTICE
    _CONVERSION_MAPPING:ClassVar[dict] = {
        "badge-info": Conversion("badge_info",str),
        "badges": Conversion("badges",lambda obj: obj.split(",")),
        "bits": Conversion("bits",str),
        "color": Conversion("color",str),
        "display-name": Conversion("display_name",str),
        "emotes": Conversion("emotes",str),
        "id": Conversion("id",str),
        "login": Conversion("login",str),
        "mod": Conversion("mod",lambda obj: bool(int(obj))),
        "msg-id": Conversion("msg_id",str),
        "room-id": Conversion("room_id",str),
        "system-msg": Conversion("system_msg",str),
        "turbo": Conversion("turbo",lambda obj: bool(int(obj))),
        "user-id": Conversion("user_id",str),
        "user-type": Conversion("user_type",str),

        "msg-param-cumulative-months":Conversion("msg_param_cumulative_months",str),
        "msg-param-displayName":Conversion("msg_param_displayName",str),
        "msg-param-login":Conversion("msg_param_login",str),
        "msg-param-months":Conversion("msg_param_months",str),
        "msg-param-promo-gift-total":Conversion("msg_param_promo_gift_total",str),
        "msg-param-promo-name":Conversion("msg_param_promo_name",str),
        "msg-param-recipient-display-name":Conversion("msg_param_recipient_display_name",str),
        "msg-param-recipient-id":Conversion("msg_param_recipient_id",str),
        "msg-param-recipient-user-name":Conversion("msg_param_recipient_user_name",str),
        "msg-param-sender-login":Conversion("msg_param_sender_login",str),
        "msg-param-sender-name":Conversion("msg_param_sender_name",str),
        "msg-param-should-share-streak":Conversion("msg_param_should_share_streak",str),
        "msg-param-streak-months":Conversion("msg_param_streak_months",str),
        "msg-param-sub-plan":Conversion("msg_param_sub_plan",str),
        "msg-param-sub-plan-name":Conversion("msg_param_sub_plan_name",str),
        "msg-param-viewerCount":Conversion("msg_param_viewerCount",str),
        "msg-param-ritual-name":Conversion("msg_param_ritual_name",str),
        "msg-param-threshold":Conversion("msg_param_threshold",str),
        "msg-param-gift-months":Conversion("msg_param_gift_months",str),
    }


# ----------------------------------------------------------------------------------------------------------------------
# - Package Imports -
# ----------------------------------------------------------------------------------------------------------------------
# General Packages
from __future__ import annotations
from dataclasses import dataclass
from typing import Literal, ClassVar
import datetime

# Athena Packages

# Local Imports
from AthenaTwitchLib.irc.tags._tags import Conversion, Tags, TAG_TYPES

# ----------------------------------------------------------------------------------------------------------------------
# - Code -
# ----------------------------------------------------------------------------------------------------------------------
@dataclass(slots=True, frozen=True)
class TagsPRIVMSG(Tags):
    """
    Class for Twitch IRC Tags, that are from the PRIVMSG message
    """
    badge_info:str=None
    badges:str=None
    bits:str=None
    color:str=None
    client_nonce:str=None
    custom_reward_id:str=None
    display_name:str=None
    emotes:str=None
    emote_only:bool=None
    first_msg:bool=None,
    flags:str=None
    id:str=None
    moderator:bool=None
    returning_chatter:bool=None
    reply_parent_msg_id:str=None
    reply_parent_user_id:str=None
    reply_parent_user_login:str=None
    reply_parent_display_name:str=None
    reply_parent_msg_body:str=None
    room_id:str=None
    subscriber:bool=None
    tmi_sent_ts:datetime.datetime=None
    turbo:bool=None
    user_id:str=None
    user_type:Literal["", "admin", "global_mod", "staff"]=None
    vip:bool=None

    _tag_type:ClassVar[TAG_TYPES] = TAG_TYPES.PRIVMSG
    _CONVERSION_MAPPING:ClassVar[dict] = {
        "badge-info": Conversion("badge_info", str),
        "badges": Conversion("badges", lambda obj: obj.split(",")),
        "bits": Conversion("bits", str),
        "color": Conversion("color", str),
        "client-nonce":Conversion("client_nonce", str),
        "custom-reward-id":Conversion("custom_reward_id", str),
        "display-name": Conversion("display_name", str),
        "emotes": Conversion("emotes", str),
        "emote-only": Conversion("emote_only", lambda obj: bool(int(obj))),
        "first-msg": Conversion("first_msg", lambda obj: bool(int(obj))),
        "flags": Conversion("flags", str),
        "id": Conversion("id", str),
        "mod": Conversion("moderator", lambda obj: bool(int(obj))),
        "returning-chatter": Conversion("returning_chatter", lambda obj: bool(int(obj))),
        "reply-parent-msg-id": Conversion("reply_parent_msg_id", str),
        "reply-parent-user-id": Conversion("reply_parent_user_id", str),
        "reply-parent-user-login": Conversion("reply_parent_user_login", str),
        "reply-parent-display-name": Conversion("reply_parent_display_name", str),
        "reply-parent-msg-body": Conversion("reply_parent_msg_body", str),
        "room-id": Conversion("room_id", str),
        "subscriber": Conversion("subscriber", lambda obj: bool(int(obj))),
        "tmi-sent-ts": Conversion("tmi_sent_ts", lambda obj: datetime.datetime.fromtimestamp(float(obj)/1000.)),
        "turbo": Conversion("turbo", lambda obj: bool(int(obj))),
        "user-id": Conversion("user_id", str),
        "user-type": Conversion("user_type", str),
        "vip": Conversion("vip", lambda obj: bool(int(obj))),
    }


# ----------------------------------------------------------------------------------------------------------------------
# - Package Imports -
# ----------------------------------------------------------------------------------------------------------------------
# General Packages
from __future__ import annotations
from dataclasses import dataclass, field
from typing import Callable
import asyncio
import functools
import socket

# Athena Packages
from AthenaLib.constants.text import NEW_LINE

# Local Imports
from AthenaTwitchLib.irc.bot import Bot
from AthenaTwitchLib.irc.data.enums import BotEvent
from AthenaTwitchLib.irc.irc_connection_protocol import IrcConnectionProtocol
from AthenaTwitchLib.logger import SectionIRC, IrcLogger

# ----------------------------------------------------------------------------------------------------------------------
# - Code -
# ----------------------------------------------------------------------------------------------------------------------
@dataclass(slots=True, kw_only=True)
class IrcConnection:
    """
    Class that constructs the IRC connection where the Bot will exist on
    The line parsing is first handled by the protocol
        to then be handed of to the bot if a command or other types of logic have been found
    """
    bot_obj:Bot

    ssl_enabled: bool = True
    irc_host: str = 'irc.chat.twitch.tv'
    irc_port: int = 6667
    irc_port_ssl: int = 6697

    protocol_cls:type[IrcConnectionProtocol]=IrcConnectionProtocol
    restart_attempts:int=5 # if set to -1, will run forever
    current_restart_attempt:int=0

    # Non init
    loop: asyncio.AbstractEventLoop = field(init=False,default_factory=asyncio.get_running_loop)
    bot_event_future: asyncio.Future = field(init=False)

    async def construct(self):
        """
        Constructor function for the Bot and all its logical systems like the asyncio.Protocol handler.
        It also logs the irc in onto the Twitch IRC server
        """
        while self.restart_attempts != 0:

            # Assemble the asyncio.Protocol
            #   The custom protocol_type needs a constructor to known which patterns to use, settings, etc...
            bot_event:asyncio.Future = self.loop.create_future()

            bot_transport, protocol_obj = await self.loop.create_connection( # type: asyncio.BaseTransport, object
                protocol_factory=self._protocol_constructor(bot_event),
                server_hostname=self.irc_host,
                ssl=self.ssl_enabled,
                sock=self._assemble_socket()
            )
            if bot_transport is None:
                IrcLogger.log_error(section=SectionIRC.CONNECTION_REFUSED)
                raise ConnectionRefusedError
            else:
                IrcLogger.log_debug(section=SectionIRC.CONNECTION_MADE)

            # Give the protocol the transporter,
            #   so it can easily create write calls to the connection
            # bot_transport: asyncio.Transport
            protocol_obj.transport = self.bot_obj.transport = bot_transport

            # Log the irc in on the IRC server
            await self.bot_obj.login()

            # noinspection PyTypeChecker
            self.bot_obj.task_logic.start_all_tasks(bot_transport, self.loop)

            # Waiting portion of the IrcConnection,
            #   This regulates the irc starting back up and restarting
            match await bot_event :
                case BotEvent.RESTART:
                    self.current_restart_attempt +=1
                    print(f"{NEW_LINE*25}{'-'*25}RESTART ATTEMPT {self.current_restart_attempt}{'-'*25}")
                    IrcLogger.log_warning(
                        section=SectionIRC.CONNECTION_RESTART,
                        text=f"attempt={self.current_restart_attempt}"
                    )

                    # Close the transport,
                    #   Else it will stay open forever
                    bot_transport.close()

                    # just wait a set interval,
                    #   to make sure we aren't seen as a ddos
                    await asyncio.sleep(0.5)

                    # Clear previous tasks
                    self.bot_obj.task_logic.stop_all_tasks()

                    # restarts it all
                    self.restart_attempts -= 1
                    continue

                case BotEvent.EXIT | _:
                    IrcLogger.log_warning(
                        section=SectionIRC.CONNECTION_EXIT,
                        text=f"called by BotEvent")
                    self.bot_obj.task_logic.stop_all_tasks()
                    self.loop.stop()
                    break

    def _assemble_socket(self) -> socket.socket:
        """
        Simple method that creates the required socket for the asyncio.Protocol factory
        """
        sock = socket.socket()
        sock.settimeout(5.)
        sock.connect(
            (
                self.irc_host,
                self.irc_port_ssl if self.ssl_enabled else self.irc_port
            )
        )
        return sock

    def _protocol_constructor(self, bot_event:asyncio.Future) -> Callable:
        """
        Simple construction for the custom asyncio.Protocol class
        Seperated into its own function for better programming common sense
        """
        return functools.partial(
            self.protocol_cls,

            # Assign the logic
            #   If this isn't defined, the protocol can't handle anything correctly
            bot_obj=self.bot_obj,

            # For restarts, exits and other special events
            bot_event_future=bot_event
        )


# ----------------------------------------------------------------------------------------------------------------------
# - Package Imports -
# ----------------------------------------------------------------------------------------------------------------------
# General Packages
from __future__ import annotations
from dataclasses import dataclass, field
from typing import Callable, ClassVar, Any
import enum

# Athena Packages
from AthenaColor import ForeNest as Fore

# Local Imports
from AthenaTwitchLib.logger import SectionIRC, IrcLogger

# ----------------------------------------------------------------------------------------------------------------------
# - Support Code -
# ----------------------------------------------------------------------------------------------------------------------
@dataclass(slots=True, frozen=True)
class Conversion:
    """
    Simple dataclass to easily hold the new attr name and type casting callback for Twitch IRC Tags
    """
    new_attr_name:str
    callback:Callable

class TAG_TYPES(enum.StrEnum):
    """
    StrEnum that holds all possible categories of Twitch IRC Tags
    """
    CLEARCHAT = enum.auto()
    CLEARMSG = enum.auto()
    GLOBALUSERSTATE = enum.auto()
    NOTICE = enum.auto()
    PRIVMSG = enum.auto()
    ROOMSTATE = enum.auto()
    USERNOTICE = enum.auto()
    USERSTATE = enum.auto()
    WHISPER = enum.auto()

    UNKNOWN = enum.auto()

# ----------------------------------------------------------------------------------------------------------------------
# - Code -
# ----------------------------------------------------------------------------------------------------------------------
@dataclass(slots=True, frozen=True)
class Tags:
    """
    Base class for all twitch IRC tag classes
    Holds some basic logic to import a tags group string, and create the correct Tags object from it
    """
    # Has to be a ClassVar,
    #   for the dataclass to know that it is a ClassVar
    _tag_type:ClassVar[TAG_TYPES] = TAG_TYPES.UNKNOWN
    _CONVERSION_MAPPING:ClassVar[dict[str:Conversion]] = {}

    @classmethod
    async def import_from_group_as_str(cls, tags:str) -> Tags:
        """
        Splits up a given tags string (eg: `badge-info=;badges=premium/1;color=#00AAAA`) into its separate tags.
        It will then cast the tags into the correct type, provided by the `cls._CONVERSION_MAPPING`
        """
        converted_tags:dict[str:Any] = {}

        for tag in tags.split(";"):
            attr_name, value = tag.split("=",1)

            if not (conversion := cls._CONVERSION_MAPPING.get(attr_name, False)):
                # If it fails, log and continue to the next one
                print(Fore.Maroon(f"TAG NAME '{attr_name}={value}' NOT FOUND IN {cls.__name__}"))
                IrcLogger.log_warning(section=SectionIRC.MSG_TAGS_UNKNOWN, text=f"{attr_name, value}")
                continue

            # When everything goes as normal
            converted_tags[conversion.new_attr_name] = conversion.callback(value)

        # noinspection PyArgumentList
        return cls(**converted_tags)



# ----------------------------------------------------------------------------------------------------------------------
# - Package Imports -
# ----------------------------------------------------------------------------------------------------------------------
# General Packages
from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Coroutine, Callable
import asyncio
import inspect

# Athena Packages

# Local Imports
from AthenaTwitchLib.irc.message_context import MessageCommandContext

# ----------------------------------------------------------------------------------------------------------------------
# - Support Code -
# ----------------------------------------------------------------------------------------------------------------------
def register_callback_as_logical_component(fnc: Callable):
    assert inspect.iscoroutinefunction(fnc), f"{fnc} was not a asyncio coroutine"
    fnc._logic_component = True

# ----------------------------------------------------------------------------------------------------------------------
# - Code -
# ----------------------------------------------------------------------------------------------------------------------
class BaseHardCodedLogic(ABC):
    """
    A class meant for hard coded tasks, commands and other logic systems handled by the bot

    Will store all functions marked as a `_logic_component` to the list of`_logic_components`
        This is useful for later parsing over these functions
    """
    _logic_components: list[Coroutine]

    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls)

        # Retrieve all items that are marked as a '_logic_component'
        obj._logic_components = list(
            filter(
                lambda i: hasattr(i, "_logic_component"),
                (getattr(obj, i) for i in obj.__dir__())
            )
        )

        return obj

# ----------------------------------------------------------------------------------------------------------------------
class BaseCommandLogic(ABC):
    """
    Logic system for commands that need to be executed.
    This is simply a base class and needs to be extended to fully work.
    """
    @abstractmethod
    async def execute_command(self, context: MessageCommandContext):
        """
        Main entry point from the Async Protocol, will first try and find a corresponding command.
        Afterwards it needs to execute the given command, with the correct context.
        """

# ----------------------------------------------------------------------------------------------------------------------
class BaseTaskLogic(ABC):
    """
    Logic system for commands that need to be executed.
    This is simply a base class and needs to be extended to fully work.
    """
    @abstractmethod
    def start_all_tasks(self, transport:asyncio.Transport, loop:asyncio.AbstractEventLoop):
        """
        Main Entry point for the constructor to start all tasks
        """

    @abstractmethod
    def stop_all_tasks(self):
        """
        Main Entry point for the constructor to stop all tasks when needed
        """


# ----------------------------------------------------------------------------------------------------------------------
# - Package Imports -
# ----------------------------------------------------------------------------------------------------------------------
# General Packages
from __future__ import annotations
from dataclasses import dataclass
from typing import ClassVar

# Athena Packages

# Local Imports
from AthenaTwitchLib.irc.tags._tags import Conversion, Tags, TAG_TYPES

# ----------------------------------------------------------------------------------------------------------------------
# - Code -
# ----------------------------------------------------------------------------------------------------------------------
@dataclass(slots=True, frozen=True)
class TagsROOMSTATE(Tags):
    """
    Class for Twitch IRC Tags, that are from the ROOMSTATE message
    """
    emote_only:bool=None
    followers_only:bool=None
    r9k:bool=None
    room_id:int=None
    slow:int=None
    subs_only:bool=None

    _tag_type:ClassVar[TAG_TYPES] = TAG_TYPES.ROOMSTATE
    _CONVERSION_MAPPING:ClassVar[dict] = {
        "emote-only": Conversion("emote_only",lambda obj: bool(int(obj))),
        "followers-only": Conversion("followers_only",lambda obj: bool(int(obj))),
        "r9k": Conversion("r9k",lambda obj: bool(int(obj))),
        "room-id": Conversion("room_id",int),
        "slow": Conversion("slow",int),
        "subs-only": Conversion("subs_only",lambda obj: bool(int(obj))),
    }


# ----------------------------------------------------------------------------------------------------------------------
# - Package Imports -
# ----------------------------------------------------------------------------------------------------------------------
# General Packages
from __future__ import annotations
from dataclasses import dataclass, asdict
import aiosqlite
import asyncio
import json

# Athena Packages
from AthenaLib.constants.types import PATHLIKE
from AthenaLib.database_connectors.async_sqlite import ConnectorAioSqlite
from AthenaLib.general.sql import sanitize_sql

# Local Imports
from AthenaTwitchLib.irc.data.enums import BotEvent, OutputTypes, CommandTypes
from AthenaTwitchLib.irc.data.sql import TBL_LOGIC_COMMANDS
from AthenaTwitchLib.irc.logic._logic import BaseCommandLogic
from AthenaTwitchLib.irc.message_context import MessageCommandContext
from AthenaTwitchLib.logger import IrcLogger, SectionIRC

# ----------------------------------------------------------------------------------------------------------------------
# - Support Code -
# ----------------------------------------------------------------------------------------------------------------------
@dataclass(slots=True, kw_only=True)
class CommandData:
    """
    Dataclass to hold the command's parameters.
    The data is gathered from the database.
    """
    id:int
    channel:str
    command_name:str
    command_arg:str|None
    command_type:str|CommandTypes
    allow_user:bool|int
    allow_sub:bool|int
    allow_vip:bool|int
    allow_mod:bool|int
    allow_broadcaster:bool
    output_text:str
    output_type:str|OutputTypes

    def __post_init__(self):
        self.command_type = CommandTypes(self.command_type)
        self.allow_user = bool(self.allow_user)
        self.allow_sub = bool(self.allow_sub)
        self.allow_vip = bool(self.allow_vip)
        self.allow_mod = bool(self.allow_mod)
        self.output_type = OutputTypes(self.output_type)

        # Log to db
        IrcLogger.log_debug(
            section=SectionIRC.CMD_DATA,
            text=json.dumps(asdict(self))
        )

# ----------------------------------------------------------------------------------------------------------------------
# - Code -
# ----------------------------------------------------------------------------------------------------------------------
class CommandLogicSqlite(BaseCommandLogic):
    """
    Logic system for commands that are retrieved from a database.
    The database in this case, is an SQLite db file.
    """
    _connector: ConnectorAioSqlite
    _special_event_mapping:dict[CommandTypes:BotEvent] = {
        CommandTypes.EXIT : BotEvent.EXIT,
        CommandTypes.RESTART: BotEvent.RESTART
    }

    def __new__(cls, *args,path:PATHLIKE, **kwargs):
        obj = super().__new__(cls, *args, *kwargs)

        # Assemble the connector
        #   Create the database as soon as possible (async function)
        obj._connector = ConnectorAioSqlite(path=path)
        asyncio.get_running_loop().create_task(obj._connector.db_create(queries=TBL_LOGIC_COMMANDS))

        return obj

    # ------------------------------------------------------------------------------------------------------------------
    # - Helper Methods -
    # ------------------------------------------------------------------------------------------------------------------
    async def first_arg_wait(self, context:MessageCommandContext, data:CommandData):
        """
        Simple method that requires the first argument fo a command to be an integer
        Will wait this amount of seconds
        """
        try:
            await asyncio.gather(
                self.output(context, data, data.output_text.format(delay=(delay := int(context.args[0])))),
                asyncio.sleep(delay)
            )
        # delay couldn't be cast into a string
        except ValueError:
            pass

    # ------------------------------------------------------------------------------------------------------------------
    # - Command execution -
    # ------------------------------------------------------------------------------------------------------------------
    async def execute_command(self, context:MessageCommandContext) -> None:
        """
        Main entry point from the Async Protocol, will first try and find a corresponding command within the database
        Afterwards it executes the command, following the correct format
        """
        # stage 1: Retrieve command
        if not (data := await self.get_command(context)): #type: CommandData
            return

        # stage 2: Validate command
        if not self.validate_user(context, data):
            return

        # stage 3: Execute command
        await self.parse_command_type(context, data)


    @staticmethod
    async def output(context: MessageCommandContext, data: CommandData, msg: str, format_:dict=None):
        """
        General output function for the Logic class to write to chat.
        """
        # Assemble the msg,
        #   as 99,99% of the time it will not fail
        msg_formatted = msg.format(
            username=context.username,
            channel=context.channel,
            **format_ if format_ is not None else {}
        )

        if data.output_type == OutputTypes.WRITE:
            await context.write(msg_formatted)
        elif data.output_type == OutputTypes.REPLY:
            await context.reply(msg_formatted)
        else:
            IrcLogger.log_error()
            raise ValueError(data.output_type)

    async def get_command(self, context:MessageCommandContext) -> CommandData|False:
        """
        Method which retrieves the command from the database, if present.
        Otherwise, will  return False.
        """
        async with self._connector.connect(commit=False) as db:

            # Prep some data so the sql query is a little easier to real
            channel = sanitize_sql(context.channel)
            cmd = sanitize_sql(context.command)
            arg = sanitize_sql(context.args[0]) if context.args else "*"

            # SQL query, might need some optimization
            async with db.execute(f"""
                SELECT * 
                FROM commands 
                WHERE (
                    (`channel`,`command_name`,`command_arg`) = ('{channel}', '{cmd}','{arg}')
                    OR (`channel`,`command_name`,`command_arg`) = ('{channel}','{cmd}','*')
                )
                ORDER BY `command_arg` DESC 
                LIMIT 1;
            """) as cursor: #type: aiosqlite.Cursor

                # Quit if nothing has been found
                if (row := await cursor.fetchone()) is None:
                    return False

                # If something has been found
                #   There are two possible states of success:
                #       - no strict args of that kind were defined
                #       - a strict arg was found and will be applied
                #   If all fails,
                #       it'll eventually return the function as false and not execute any further
                match context, data := CommandData(**dict(row)):
                    case MessageCommandContext(args=_), CommandData(command_arg=None|"*"):
                        return data

                    case MessageCommandContext(args=args), CommandData(command_arg=stored_arg) if stored_arg == args[0]:
                        return data

        # Nothing matched
        return False

    @staticmethod
    def validate_user(context:MessageCommandContext, data:CommandData) -> bool:
        """
        Method checks if the user can use the command
        """
        # Any will return false if nothing checks
        #   True if at least one is successful
        return any((
            data.allow_user,
            data.allow_broadcaster and context.user == f":{context.channel}!{context.channel}@{context.channel}.tmi.twitch.tv",
            data.allow_mod and context.tags.moderator,
            data.allow_sub and context.tags.subscriber,
            data.allow_vip and context.tags.vip,
        ))

    async def parse_command_type(self, context:MessageCommandContext, data:CommandData) -> None:
        """
        If a command has been found, parse it's args and text corresponding to the CommandType
        """
        match data:
            case CommandData(command_type=CommandTypes.DEFAULT):
                await self.output(context, data, data.output_text)

            case CommandData(command_type=CommandTypes.ARGS) if context.args:
                await self.output( context, data, data.output_text, format_={f"args_{i}":a for i, a in enumerate(context.args)})

            # Special type of command
            #   Like the exit or restart commands
            case CommandData(command_type=cmd_type) if cmd_type in (CommandTypes.EXIT,CommandTypes.RESTART):
                if context.args:
                    await self.first_arg_wait(context=context,data=data)

                # Set the event future to the correct event
                #   constructor will handle the rest
                context.bot_event_future.set_result(self._special_event_mapping.get(cmd_type, None))



# ----------------------------------------------------------------------------------------------------------------------
# - Package Imports -
# ----------------------------------------------------------------------------------------------------------------------
# General Packages
from __future__ import annotations
from dataclasses import dataclass
from typing import Callable
import asyncio
import datetime

# Athena Packages

# Local Imports
from AthenaTwitchLib.irc.logic._logic import BaseHardCodedLogic, BaseTaskLogic, register_callback_as_logical_component


# ----------------------------------------------------------------------------------------------------------------------
# - Code -
# ----------------------------------------------------------------------------------------------------------------------
@dataclass(slots=True, kw_only=True)
class TaskData:
    """
    Simple dataclass to hold basic information about the hard coded task.
    Meant for the `TaskLogic` class to differentiate what to do when it needs to execute the tasks
    """
    at:datetime.timedelta = None
    interval:datetime.timedelta = None
    channel:str = None

    def __post_init__(self):
        if self.at is not None:
            self.interval = self.at

# ----------------------------------------------------------------------------------------------------------------------
# - Code -
# ----------------------------------------------------------------------------------------------------------------------
class TaskLogic(BaseHardCodedLogic,BaseTaskLogic):
    """
    Logic system behind hard coded tasks.
    """
    open_tasks:list[asyncio.Task]
    _tasks:list[tuple[Callable,TaskData]]

    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls, *args, *kwargs)

        # retrieve all tasks
        obj._tasks = [
            (coroutine, coroutine._data)
            for coroutine in
            obj._logic_components
        ]

        # Extract data from the bot, and make sure that tasks adhere to this

        obj.open_tasks = []
        return obj

    def start_all_tasks(self, transport:asyncio.Transport, loop:asyncio.AbstractEventLoop):
        if self.open_tasks:
            self.stop_all_tasks()

        self.open_tasks = [
            loop.create_task(self._create_task(transport, coroutine, task_data))
            for coroutine, task_data in self._tasks
        ]

    def stop_all_tasks(self):
        for task in self.open_tasks:
            task.cancel()
        self.open_tasks.clear()

    async def _create_task(self, transport:asyncio.Transport, coroutine:Callable, task_data:TaskData):
        """
        Main function of a task
        Gets called by `TaskLogic.start_all_tasks` as it controls the actions of a task
        """

        # it the time has been set to be at a certain part of the hour
        #   Wait until we get to that part of the hour
        if task_data.at is not None:
            now = datetime.datetime.now()
            await asyncio.sleep((
                task_data.at
                - (datetime.timedelta(hours=now.hour, minutes=now.minute, seconds=now.second) % task_data.at)
            ).seconds)

            while True:
                await coroutine(transport)
                await asyncio.sleep(task_data.interval.total_seconds())

        # If we don't have to set the task at a certain part of the hour
        #   Start up the loop regularly
        else:
            while True:
                await asyncio.sleep(task_data.interval.total_seconds())
                await coroutine(transport)


    @staticmethod
    def task(task_data:TaskData):
        """
        Decorator to be used by the Bot,
            to assign a method as a task that has to be run at a given interval
        """
        def decorator(fnc):
            register_callback_as_logical_component(fnc)
            fnc._data = task_data
            return fnc
        return decorator


# ----------------------------------------------------------------------------------------------------------------------
# - Package Imports -
# ----------------------------------------------------------------------------------------------------------------------
# General Packages
from __future__ import annotations
from dataclasses import dataclass, field
from typing import Any, Callable
import asyncio
import functools
import json
import re

# Athena Packages
from AthenaColor import ForeNest as Fore
from AthenaLib.general.json import GeneralCustomJsonEncoder

# Local Imports
from AthenaTwitchLib.irc.bot import Bot
from AthenaTwitchLib.irc.data.enums import BotEvent
from AthenaTwitchLib.irc.message_context import MessageContext,MessageCommandContext
from AthenaTwitchLib.irc.tags import TagsPRIVMSG, TagsUSERSTATE, TagsUSERNOTICE
from AthenaTwitchLib.logger import SectionIRC, IrcLogger
import AthenaTwitchLib.irc.data.regex as RegexPatterns

# ----------------------------------------------------------------------------------------------------------------------
# - Support Code -
# ----------------------------------------------------------------------------------------------------------------------
class _TransportBuffer:
    """
    Simple class to be used by the `IrcConnectionProtocol`, before the transporter object is actually set.
    It keeps the to be sent message into a small buffer, for it to then be parsed and deleted once the
    actual transporter is present in the `IrcConnectionProtocol`
    """
    buffer: list[bytes] = []

    @classmethod
    def write(cls, data:bytes):
        """
        Stores data to the buffer
        """
        cls.buffer.append(data)

def log_handler(fnc:Callable) -> Any:
    """
    Simple decorator to keep track of how many calls are made to handlers
    """
    @functools.wraps(fnc)
    async def wrapper(*args, **kwargs):
        IrcLogger.log_track(section=SectionIRC.HANDLER_CALLED, text=fnc.__name__),
        if (line := kwargs.get("line", None)) is not None:
            IrcLogger.log_debug(section=SectionIRC.MSG_ORIGINAL, text=line),

        return await fnc(*args, **kwargs)

    return wrapper

# ----------------------------------------------------------------------------------------------------------------------
# - Code -
# ----------------------------------------------------------------------------------------------------------------------
@dataclass(slots=True)
class IrcConnectionProtocol(asyncio.Protocol):
    """
    Asyncio.Protocol child class,
    Holds all logic to convert the incoming Twitch IRC messages to useful calls/data
    """
    bot_event_future: asyncio.Future
    bot_obj:Bot

    _transport: asyncio.transports.Transport = None  # delayed as it has to be set after the connection has been made
    loop :asyncio.AbstractEventLoop = field(init=False)

    def __post_init__(self):
        self.loop = asyncio.get_running_loop()

    # ------------------------------------------------------------------------------------------------------------------
    # - Properties -
    # ------------------------------------------------------------------------------------------------------------------
    @property
    def transport(self):
        """
        Getter of the `transport` property
        This is necessary as transport is set later by the constructor than when the protocol is created
        When the transport isn't set yet, it will store write data to a temp buffer
        This buffer will be removed after the setter of the `transport` property is called
        """
        if self._transport is None:
            return _TransportBuffer

        return self._transport

    @transport.setter
    def transport(self, value:asyncio.transports.Transport):
        """
        Setter of the `transport` property
        Executes any write calls in the buffer, and the deletes the buffer
        """
        self._transport = value

        if _TransportBuffer.buffer:
            for data in _TransportBuffer.buffer:
                self._transport.write(data)

            _TransportBuffer.buffer.clear()

    # ------------------------------------------------------------------------------------------------------------------
    # - Protocol Calls (aka, calls made by asyncio.Protocol) -
    # ------------------------------------------------------------------------------------------------------------------
    def data_received(self, data: bytearray) -> None:
        """
        First hit of the protocol when it receives data from Twitch IRC
        Because twitch sends in this data in bytes, and sometimes multiple different message,
        the function has to decode and split the data on every new line
        """

        # TODO sort on most used messages

        for line in data.decode().split("\r\n"):
            # An Empty line
            if not line:
                continue

            elif message := RegexPatterns.message.match(line):
                if (
                    (cmd_match := RegexPatterns.message_command.match(message.groups()[-1]) )
                    and cmd_match.groups()[0] == self.bot_obj.prefix
                ):
                    self.loop.create_task(self.handle_message_command(message, cmd_match, line=line))
                else:
                    self.loop.create_task(self.handle_message(message, line=line))

            elif line == "PING :tmi.twitch.tv":
                self.loop.create_task(self.handle_ping(line=line))

            elif (
                (server_message := RegexPatterns.server_message.match(line))
                and server_message.groups()[0] == self.bot_obj.name
            ):
                self.loop.create_task(self.handle_server_message(server_message, line=line))

            elif join := RegexPatterns.join.match(line):
                self.loop.create_task(self.handle_join(join, line=line))

            elif part := RegexPatterns.part.match(line):
                self.loop.create_task(self.handle_part(part, line=line))

            elif server_353 := RegexPatterns.server_353.match(line):
                self.loop.create_task(self.handle_server_353(server_353, line=line))

            elif server_366 := RegexPatterns.server_366.match(line):
                self.loop.create_task(self.handle_server_366(server_366, line=line))

            elif server_cap := RegexPatterns.server_cap.match(line):
                self.loop.create_task(self.handle_server_cap(server_cap, line=line))

            elif user_notice := RegexPatterns.user_notice.match(line):
                self.loop.create_task(self.handle_user_notice(user_notice, line=line))

            elif user_notice_raid := RegexPatterns.user_notice_raid.match(line):
                self.loop.create_task(self.handle_user_notice_raid(user_notice_raid, line=line))

            elif user_state := RegexPatterns.user_state.match(line):
                self.loop.create_task(self.handle_user_state(user_state, line=line))

            else:
                self.loop.create_task(self.handle_UNKNOWN(line))

    def connection_lost(self, exc: Exception | None) -> None:
        if exc is not None:
            print(exc)

        if not self.bot_event_future.done():
            self.bot_event_future.set_result(BotEvent.RESTART)

    # ------------------------------------------------------------------------------------------------------------------
    # - Line handlers -
    # ------------------------------------------------------------------------------------------------------------------
    @log_handler
    async def handle_ping(self,*, line):
        """
        Method is called when the Twitch server sends a keep alive PING message
        Needs to have the reply: `"PONG :tmi.twitch.tv` for the connection to remain alive
        """
        print(f"{Fore.Peru('PONG')} | {line}")

        # Need to keep alive
        self.transport.write("PONG :tmi.twitch.tv\r\n".encode())

    # ------------------------------------------------------------------------------------------------------------------
    @log_handler
    async def handle_server_message(self, server_message:re.Match, *, line:str):
        """
        Method is called when the Twitch server sends a message that isn't related to any user or room messages
        """
        print(f"{Fore.Blue('SERVER_MESSAGE')} | {line}")

    # ------------------------------------------------------------------------------------------------------------------
    @log_handler
    async def handle_server_353(self, server_353: re.Match, *, line: str):
        """
        Method is called when twitch sends a 353 message
        """
        print(f"{Fore.AliceBlue('SERVER_353')} | {line}")

    # ------------------------------------------------------------------------------------------------------------------
    @log_handler
    async def handle_server_366(self, server_366: re.Match, *, line: str):
        """
        Method is called when twitch sends a 353 message
        """
        print(f"{Fore.Ivory('SERVER_366')} | {line}")

    # ------------------------------------------------------------------------------------------------------------------
    @log_handler
    async def handle_server_cap(self, server_cap: re.Match, *, line: str):
        """
        Method is called when twitch sends a CAP message
        """
        print(f"{Fore.Khaki('SERVER_CAP')} | {line}")

    # ------------------------------------------------------------------------------------------------------------------
    @log_handler
    async def handle_join(self, join:re.Match,*, line:str):
        """
        Method is called when any user (irc or viewer) joins the channel
        """
        print(f"{Fore.Red('JOIN')} | {line}")

    # ------------------------------------------------------------------------------------------------------------------
    @log_handler
    async def handle_part(self, part:re.Match, *, line:str):
        """
        Method is called when any user (irc or viewer) parts the channel
        """
        print(f"{Fore.DeepPink('PART')} | {line}")

    # ------------------------------------------------------------------------------------------------------------------
    @log_handler
    async def handle_message(self, message:re.Match, *, line:str):
        """
        Method is called when any user (irc or viewer) sends a regular message in the channel
        """
        print(f"{Fore.Orchid('MESSAGE')} | {message.groups()[-1]} | {Fore.SlateGray(line)}")

        # Extract data from matched message
        #   Easily done due to regex groups
        tags_group_str,user,channel,text = message.groups()

        # Create the context and run more checks
        message_context = MessageContext(
            tags=await TagsPRIVMSG.import_from_group_as_str(tags_group_str),
            user=user,
            username=RegexPatterns.username.findall(user)[0],
            channel=channel,
            text=text,
            transport=self.transport,
            bot_event_future=self.bot_event_future,
            original_line=line
        )
        IrcLogger.log_debug(
            section=SectionIRC.MSG_CONTEXT,
            text=json.dumps(message_context.as_dict(), cls=GeneralCustomJsonEncoder)
        )

    # ------------------------------------------------------------------------------------------------------------------
    @log_handler
    async def handle_message_command(self, message:re.Match, cmd_match:re.Match, *, line:str):
        """
        Method is called when any user (irc or viewer) sends a message in the channel,
        which is presumed to be an irc command
        """
        print(f"{Fore.Orchid('MESSAGE_COMMAND')} | {message.groups()[-1]} | {Fore.SlateGray(line)}")

        # Extract data from matched message
        #   Easily done due to regex groups
        tags_group_str,user,channel,text = message.groups()

        # prefix, command, args
        _,command, args = cmd_match.groups()

        message_context = MessageCommandContext(
            tags=await TagsPRIVMSG.import_from_group_as_str(tags_group_str),
            user=user,
            username=RegexPatterns.username.findall(user)[0],
            channel=channel,
            text=f"!{command}",
            transport=self.transport,
            bot_event_future=self.bot_event_future,
            original_line=line,
            command=command,
            args=args.strip().split(" ")
        )

        IrcLogger.log_debug(
            section=SectionIRC.MSG_CONTEXT,
            text=json.dumps(message_context.as_dict(), cls=GeneralCustomJsonEncoder)
        )

        await self.bot_obj.command_logic.execute_command(
            context=message_context
        )

    # ------------------------------------------------------------------------------------------------------------------
    @log_handler
    async def handle_user_notice(self, user_notice:re.Match, *, line:str):
        """
        Method is called when twitch sends a USERNOTICE message
        """
        tags_group_str,user,channel,text = user_notice.groups()
        tags = await TagsUSERNOTICE.import_from_group_as_str(tags_group_str)
        print(f"{Fore.Plum('USERNOTICE')} | {line}")

    # ------------------------------------------------------------------------------------------------------------------
    @log_handler
    async def handle_user_notice_raid(self, user_notice_raid:re.Match, *, line:str):
        """
        Method is called when twitch sends a USERNOTICE message
        """
        tags_group_str,channel = user_notice_raid.groups()
        tags = await TagsUSERNOTICE.import_from_group_as_str(tags_group_str)
        print(f"{Fore.Plum('USERNOTICE RAID')} | {line} | {tags}")

    # ------------------------------------------------------------------------------------------------------------------
    @log_handler
    async def handle_user_state(self, user_state:re.Match, *, line:str):
        """
        Method is called when twitch sends a USERSTATE message
        """
        tags_group_str,channel = user_state.groups()
        tags = await TagsUSERSTATE.import_from_group_as_str(tags_group_str)
        print(f"{Fore.Plum('USERSTATE')} | {line} | {tags}")

    # ------------------------------------------------------------------------------------------------------------------
    @log_handler
    async def handle_UNKNOWN(self, line:str):
        """
        Method is called when the protocol can't find an appropriate match for the given string
        """
        print(Fore.SlateGray(f"NOT CAUGHT | {line}"))
        IrcLogger.log_warning(
            section=SectionIRC.HANDLER_UNKNOWN,
            text=line
        )